from typing import Any

# https://docs.python.org/3/reference/datamodel.html#object.__new__


class Meta(type):
#     def __new__(mcs, name, bases, namespace):
#         return super().__new__(mcs, name, bases, namespace)
#     
#     def __init__(cls, name, bases, namespace):
#         super().__init__(name, bases, namespace)
#     
#     def __call__(cls, *args, **kwargs):
#         super().__call__(*args, **kwargs)
#     
#     def __str__(cls) -> str:
#         super().__str__()
#     
#     def __repr__(cls) -> str:
#         super().__repr__()
#     
#     def __hash__(cls) -> int:
#         super().__hash__()
#     
#     def __eq__(cls, other) -> bool:
#         super().__eq__(other)
#     
#     def __ne__(cls, other) -> bool:
#         super().__ne__(other)
#     
#     def __lt__(cls, other) -> bool:
#         super().__lt__(other)
#     
#     def __le__(cls, other) -> bool:
#         super().__le__(other)
#     
#     def __gt__(cls, other) -> bool:
#         super().__gt__(other)
#     
    def __getattribute__(cls, name: str) -> Any:
        return super().__getattribute__(name)

    def __setattr__(cls, name: str, value: Any) -> None:
        super().__setattr__(name, value)

    def __delattr__(cls, name: str) -> None:
        super().__delattr__(name)

    def __dir__(cls) -> list[str]:
        super().__dir__()
    
#     def __instancecheck__(cls, instance: Any) -> bool:
#         super().__instancecheck__(instance)
#     def __subclasscheck__(cls, subclass: Any) -> bool:
#         super().__subclasscheck__(subclass)

class PersistentObject(metaclass=Meta):
    session_state = None

    def __getattribute__(self, name: str):
        session_state = super().__getattribute__("session_state")
        if name in list(session_state.keys()):
            # Get (user-defined) value from a key in the session state
            return session_state[name]
        else:
            # Get class attribute
            return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name not in super().__dir__():
            # Set (user-defined) value at a key in the session state
            session_state = super().__getattribute__("session_state")
            session_state[name] = value
        else:
            # Set class attribute
            super().__setattr__(name, value)

    def __delattr__(self, name: str) -> None:
        if name not in super().__dir__():
            # Delete (user-defined) value at a key in the session state
            session_state = super().__getattribute__("session_state")
            del session_state[name]
        else:
            # Delete class attribute
            super().__delattr__(name)

    def __dir__(self) -> list[str]:
        session_state = super().__getattribute__("session_state")
        return super().__dir__() + list(session_state.keys())
